#include <microfx/microfx.h>
#include <microfx/ext/strtools.h>

#define GATE_SIZE 16
#define NUMBER_OF_BALLS_IN_FIRST_LEVEL 10
#define NUMBER_OF_BALLS_ADDED_EACH_LEVEL 4


typedef struct{
	short x, y;
	char dx, dy, color;
} Ball;

/* srand and rand from https://stackoverflow.com/questions/4768180/rand-implementation */
static unsigned long int next = 1;

int rand(void) // RAND_MAX assumed to be 32767
{
    next = next * 1103515245 + 12345;
    return (unsigned int)(next/65536) % 32768;
}

void srand(unsigned int seed)
{
    next = seed;
}
/********************/

int abs(int n) {
    return n<0 ? -n : n;
}

int time(void) {
	MRtc rtc;
	rgettime(&rtc);
	return rtc.hour*3600 + rtc.minute*60 + rtc.second;
}

char* int2str(char* c, unsigned int n)
{
    if(n==0)
    {
        c[0] = '0';
        c[1] = 0;
    }
    else
    {
        int i, l=0;
        for(i=n ; i ; i/=10)
            l++;
        c[l] = 0;
        for(i=n ; i ; i/=10)
            c[--l] = i%10+'0';
    }
    return c;
}

char* writeTime(char* str, int t)
{
	char c[10];
	unsigned char h, m, s;
	if(t<0) t += 24*3600;
	str[0] = 0;
	h = t/3600;
	t %=3600;
	m = t/60;
	s = t%60;
	if(h) {
		int2str(str, h);
		strcat(str, "h");
	}
	if(m || h) {
		strcat(str, int2str(c, m));
		strcat(str, "m");
	}
	strcat(str, int2str(c, s));
	strcat(str, "s");
	return str;
}

void newBall(Ball* ball)
{
	do { ball->x = rand()%1220+30; } while(ball->x>57 && ball->x<70);
	ball->y = rand()%580+30;
	do { ball->dx = rand()%11-5; } while(abs(ball->dx)<3);
	do { ball->dy = rand()%11-5; } while(abs(ball->dy)<3);
	ball->color = rand()%2;
}

void initBalls(Ball** ball, unsigned int *n)
{
	int i;
	if(*ball) free(*ball);
	*ball = malloc(*n * sizeof(Ball));
	while(*ball == NULL) {
		*n--;
		*ball = malloc(*n * sizeof(Ball));
	}
	for(i=0 ; i<*n ; i++)
		newBall(&(*ball)[i]);
}

void drawBall(Ball b)
{
	unsigned char bmp[4] = {96, 240, 240, 96};
	char *vram, *p, i, lag, x, y;
	unsigned short line, byte;
	x = b.x/10;
	y = b.y/10;
	p = (char*)&line;
	vram = sgetvram();
	lag = 8-x%8;
	byte = y*16 + x/8;
	for(i=0 ; i<4 ; i++)
	{
		line = bmp[i] << lag;
		vram[byte] |= *p;
		vram[byte+1] |= *(p+1);
		byte += 16;
	}
	if(b.color)
	{
		byte = (y+1)*16 + x/8;
		line = ~line;
		for(i=0 ; i<2 ; i++)
		{
			vram[byte] &= *p;
			vram[byte+1] &= *(p+1);
			byte += 16;
		}
	}
}

int main(void)
{
	int i, y = 32-GATE_SIZE/2, time_begin, time_end, max_ball=NUMBER_OF_BALLS_IN_FIRST_LEVEL;
	char win = 0, str[20];
	Ball *ball = NULL;

	srand(time_begin = time());
	initBalls(&ball, &max_ball);

    while(!kcheck(KCEXIT))
    {
		if(kcheck(KCUP) && y>2) y--;
		if(kcheck(KCDOWN) && y<62-GATE_SIZE) y++;

/* balls movements */
		for(i=0 ; i<max_ball ; i++)
		{
			ball[i].x += ball[i].dx;
			ball[i].y += ball[i].dy;
			if(ball[i].x < 20)
			{
				ball[i].x = 20;
				ball[i].dx = abs(ball[i].dx);
			}
			if(ball[i].x > 1220)
			{
				ball[i].x = 1220;
				ball[i].dx = -abs(ball[i].dx);
			}
			if(ball[i].y < 20)
			{
				ball[i].y = 20;
				ball[i].dy = abs(ball[i].dy);
			}
			if(ball[i].y > 580)
			{
				ball[i].y = 580;
				ball[i].dy = -abs(ball[i].dy);
			}
			if(ball[i].y < y*10 || ball[i].y+30 > 10*(y+GATE_SIZE-1) || win)
			{
				if(ball[i].x/10==58) ball[i].dx = -abs(ball[i].dx);
				if(ball[i].x/10==66) ball[i].dx = abs(ball[i].dx);
			}
			if(ball[i].x>580 && ball[i].x<660)
			{
				if(ball[i].y <= y*10) ball[i].dy = abs(ball[i].dy);
				if(ball[i].y+30 >= 10*(y+GATE_SIZE-1)) ball[i].dy = -abs(ball[i].dy);
			}
		}

/* test if win */
		if(!win) win = 1;
		for(i=0 ; i<max_ball ; i++)
		{
			if(ball[i].color)
			{
				if(ball[i].x < 650)
					win = 0;
			} else {
				if(ball[i].x > 590)
					win = 0;
			}
		}

/* Drawing */
    	sclear();
    	sline(1, 1, 1, 62, SBLACK); //left
    	sline(126, 1, 126, 62, SBLACK); //right
    	sline(1, 1, 62, 1, SBLACK); //top left
    	sline(126, 1, 65, 1, SBLACK); //top right
    	sline(1, 62, 62, 62, SBLACK); //bottom left
    	sline(126, 62, 65, 62, SBLACK); //bottom right
    	if(win)
    	{
    		sline(62, 1, 62, 62, SBLACK);
    		sline(65, 1, 65, 62, SBLACK);
    	} else {
			sline(62, 1, 62, y-1, SBLACK); //middle top left
			sline(65, 1, 65, y-1, SBLACK); //middle top right
			sline(62, y-1, 65, y-1, SBLACK); //middle top bottom
			sline(62, 62, 62, y+GATE_SIZE, SBLACK); //middle bottom left
			sline(65, 62, 65, y+GATE_SIZE, SBLACK); //middle bottom right
			sline(62, y+GATE_SIZE, 65, y+GATE_SIZE, SBLACK); //middle bottom top
    	}

    	sline(0, 0, 0, 63, SBLACK); //border left
    	sline(0, 0, 60, 0, SBLACK); //border top
    	sline(0, 63, 60, 63, SBLACK); //border bottom

    	for(i=0 ; i<max_ball ; i++)
    		drawBall(ball[i]);

		if(win)
		{
			if(win==1)
			{
				time_end = time();
				writeTime(str, time_end-time_begin);
			}
			if(win > 100)
			{
				gpopup(3);
				slocate(3, 3, "You won in");
				slocate(14, 3, str);
				slocate(3, 4, "Do you want more?");
				slocate(7, 5, "Press EXE");
				if(kcheck(KCEXE))
				{
					max_ball += NUMBER_OF_BALLS_ADDED_EACH_LEVEL;
					initBalls(&ball, &max_ball);
					time_begin = time();
				}
			}
			else win++;
		}
		supdate();
		tsleep_ms(10);
    }
    return 1;
}

